import locale
import logging
import math
import os

import numpy as np
from PIL import Image, ImageFilter

from pipictureframe.picdb.PictureUpdater import PictureData

log = logging.getLogger(__name__)

try:
    import pi3d
    from pi3d.Texture import MAX_SIZE
except ModuleNotFoundError as e:
    if "pygame" in e.msg:
        log.error(
            "Pygame must be installed manually if running on Windows. "
            'Please run "pip install pygame".'
        )
        exit(2)
    else:
        raise e


class Pi3dFuncs:
    def __init__(self, config):
        # noinspection PyBroadException
        try:
            locale.setlocale(locale.LC_TIME, config.local)
        except Exception as e:
            log.warning("error trying to set local to {}".format(config.local), e)

        self._display = pi3d.Display.create(
            x=config.display_x,
            y=config.display_y,
            w=config.display_w,
            h=config.display_h,
            frames_per_second=config.fps,
            display_config=pi3d.DISPLAY_CONFIG_HIDE_CURSOR,
            background=config.background,
        )
        self._camera = pi3d.Camera(is_3d=False)

        shader = pi3d.Shader(config.shader)
        self._slide = pi3d.Sprite(
            camera=self._camera, w=self._display.width, h=self._display.height, z=5.0
        )
        self._slide.set_shader(shader)
        self._slide.unif[47] = config.edge_alpha
        self._slide.unif[54] = config.blend_type
        self._slide.unif[55] = 1.0  # brightness used by shader [18][1]

        if config.keyboard:
            self.keyboard = pi3d.Keyboard()

        grid_size = math.ceil(len(config.codepoints) ** 0.5)
        font = pi3d.Font(
            config.font_file, codepoints=config.codepoints, grid_size=grid_size
        )
        self._text = pi3d.PointText(
            font, self._camera, max_chars=200, point_size=config.show_text_sz
        )
        self._textblock = pi3d.TextBlock(
            x=0,
            y=-self._display.height // 2 + (config.show_text_sz // 2) + 20,
            z=0.1,
            rot=0.0,
            char_count=199,
            text_format="{}".format(" "),
            size=0.99,
            spacing="F",
            justify=0.5,
            space=0.02,
            colour=(1.0, 1.0, 1.0, 1.0),
        )
        self._text.add_text_block(self._textblock)

        bkg_ht = self._display.height // 4
        text_bkg_array = np.zeros((bkg_ht, 1, 4), dtype=np.uint8)
        text_bkg_array[:, :, 3] = np.linspace(0, 170, bkg_ht).reshape(-1, 1)
        text_bkg_tex = pi3d.Texture(text_bkg_array, blend=True, free_after_load=True)
        self._text_bkg = pi3d.Plane(
            w=self._display.width,
            h=bkg_ht,
            y=-self._display.height // 2 + bkg_ht // 2,
            z=4.0,
        )
        back_shader = pi3d.Shader("uv_flat")
        self._text_bkg.set_draw_details(back_shader, [text_bkg_tex])

        self._foreground_slide = None
        self._background_slide = None

        # return display, camera, slide, keyboard, text_block, text, text_bkg

    @staticmethod
    def orientate_image(im, orientation):
        if orientation == 2:
            im = im.transpose(Image.FLIP_LEFT_RIGHT)
        elif orientation == 3:
            im = im.transpose(Image.ROTATE_180)  # rotations are clockwise
        elif orientation == 4:
            im = im.transpose(Image.FLIP_TOP_BOTTOM)
        elif orientation == 5:
            im = im.transpose(Image.FLIP_LEFT_RIGHT).transpose(Image.ROTATE_270)
        elif orientation == 6:
            im = im.transpose(Image.ROTATE_270)
        elif orientation == 7:
            im = im.transpose(Image.FLIP_LEFT_RIGHT).transpose(Image.ROTATE_90)
        elif orientation == 8:
            im = im.transpose(Image.ROTATE_90)
        return im

    def texture_load(self, picture: PictureData, config):
        size = (self._display.width, self._display.height)
        pic_path = picture.absolute_path
        orientation = picture.orientation

        try:
            im = Image.open(pic_path)

            (w, h) = im.size
            # changing MAX_SIZE causes serious crash on linux laptop!
            max_dimension = MAX_SIZE
            # turned off for 4K display - will cause issues on RPi before v4
            if not config.auto_resize:
                max_dimension = 3840  # TODO check if mipmapping should be turned off with this setting.
            if w > max_dimension:
                im = im.resize(
                    (max_dimension, int(h * max_dimension / w)), resample=Image.BICUBIC
                )
            elif h > max_dimension:
                im = im.resize(
                    (int(w * max_dimension / h), max_dimension), resample=Image.BICUBIC
                )
            if orientation > 1:
                im = Pi3dFuncs.orientate_image(im, orientation)
            if config.blur_edges and size is not None:
                wh_rat = (size[0] * im.height) / (size[1] * im.width)
                # make a blurred background if wh_rat is not exactly 1
                if abs(wh_rat - 1.0) > 0.01:
                    (sc_b, sc_f) = (size[1] / im.height, size[0] / im.width)
                    if wh_rat > 1.0:
                        (sc_b, sc_f) = (sc_f, sc_b)  # swap round
                    (w, h) = (
                        round(size[0] / sc_b / config.blur_zoom),
                        round(size[1] / sc_b / config.blur_zoom),
                    )
                    (x, y) = (round(0.5 * (im.width - w)), round(0.5 * (im.height - h)))
                    box = (x, y, x + w, y + h)
                    blr_sz = (int(x * 512 / size[0]) for x in size)
                    im_b = im.resize(size, resample=0, box=box).resize(blr_sz)
                    im_b = im_b.filter(ImageFilter.GaussianBlur(config.blur_amount))
                    im_b = im_b.resize(size, resample=Image.BICUBIC)
                    im_b.putalpha(
                        round(255 * config.edge_alpha)
                    )  # to apply the same EDGE_ALPHA as the no blur method.
                    im = im.resize(
                        (int(x * sc_f) for x in im.size), resample=Image.BICUBIC
                    )
                    """resize can use Image.LANCZOS (alias for Image.ANTIALIAS) for resampling
                    for better rendering of high-contranst diagonal lines. NB downscaled large
                    images are rescaled near the start of this try block if w or h > max_dimension
                    so those lines might need changing too.
                    """
                    im_b.paste(
                        im,
                        box=(
                            round(0.5 * (im_b.width - im.width)),
                            round(0.5 * (im_b.height - im.height)),
                        ),
                    )
                    im = im_b  # have to do this as paste applies in place
            tex = pi3d.Texture(
                im,
                blend=True,
                m_repeat=True,
                automatic_resize=config.auto_resize,
                free_after_load=True,
            )
            # tex = pi3d.Texture(im, blend=True, m_repeat=True, automatic_resize=config.AUTO_RESIZE,
            #                    mipmap=config.AUTO_RESIZE, free_after_load=True)
            # poss try this if still some artifacts with full resolution
        except Exception as e:
            log.error(f"Error loading file {pic_path}", e)
            tex = None
        return tex

    def load_fg(self, cur_pic: PictureData, config):
        self._foreground_slide = self.texture_load(cur_pic, config)
        wh_rat = (self._display.width * self._foreground_slide.iy) / (
            self._display.height * self._foreground_slide.ix
        )
        if (wh_rat > 1.0 and config.fit) or (wh_rat <= 1.0 and not config.fit):
            sz1, sz2, os1, os2 = 42, 43, 48, 49
        else:
            sz1, sz2, os1, os2 = 43, 42, 49, 48
            wh_rat = 1.0 / wh_rat
        self._slide.unif[sz1] = wh_rat
        self._slide.unif[sz2] = 1.0
        self._slide.unif[os1] = (wh_rat - 1.0) * 0.5
        self._slide.unif[os2] = 0.0

    def copy_fg_to_bg(self):
        self._background_slide = self._foreground_slide
        self._slide.unif[45:47] = self._slide.unif[
            42:44
        ]  # transfer front width and height factors to back
        self._slide.unif[51:53] = self._slide.unif[
            48:50
        ]  # transfer front width and height offsets

    # --- Sanitize the specified string by removing any chars not found in codepoints
    @staticmethod
    def sanitize_string(string, codepoints):
        return "".join([c for c in string if c in codepoints])

    @staticmethod
    def smooth_alpha(a):
        return a * a * (3.0 - 2.0 * a)

    def set_fg_bg_ratio(self, ratio):
        """
        Set ratio of foreground to background. 0 = only bg, 1 = only fg.
        """
        self._slide.unif[44] = self.smooth_alpha(ratio)

    def prepare_text(self, cur_pic: PictureData, config):
        string_components = []
        if config.show_text > 0:
            if (config.show_text & 1) == 1:  # name
                string_components.append(
                    self.sanitize_string(
                        os.path.basename(cur_pic.absolute_path), config.codepoints
                    )
                )
            if (config.show_text & 2) == 2:  # date
                string_components.append(
                    cur_pic.get_date_time().strftime(config.show_date_fm)
                )
            # TODO fix geoloc
            # if config.LOAD_GEOLOC and (config.SHOW_TEXT & 4) == 4:  # location
            #    loc_string = sanitize_string(iFiles[pic_num].location.strip())
            #    if loc_string:
            #        string_components.append(loc_string)
            if (config.show_text & 8) == 8:  # folder
                string_components.append(
                    self.sanitize_string(
                        os.path.basename(os.path.dirname(cur_pic.absolute_path)),
                        config.codepoints,
                    )
                )

            final_string = " • ".join(string_components)
            self._textblock.set_text(text_format=final_string, wrap=config.text_width)

            last_ch = len(final_string)
            adj_y = (
                self._text.locations[:last_ch, 1].min() + self._display.height // 2
            )  # y pos of last char rel to bottom of screen
            self._textblock.set_position(
                y=(self._textblock.y - adj_y + config.show_text_sz)
            )

        else:
            self._textblock.set_text(text_format=" ")
            self._textblock.colouring.set_colour(alpha=0.0)
            self._text_bkg.set_alpha(0.0)
        self._text.regen()

    def set_text_alpha_and_draw(self, alpha):
        alpha = self.smooth_alpha(alpha)
        self._textblock.colouring.set_colour(alpha=alpha)
        self._text.regen()
        self._text_bkg.set_alpha(alpha)
        self._text_bkg.draw()
        self._text.draw()

    def set_textures(self):
        self._slide.set_textures([self._foreground_slide, self._background_slide])

    def display_loop_running(self) -> bool:
        return self._display.loop_running()

    def draw_slide(self):
        self._slide.draw()



import logging
import random
import time
from datetime import datetime
from typing import List, Optional
from multiprocessing import Process

from pipictureframe.picdb.Database import (
    Database,
    LAST_DB_UPDATE_KEY_STR,
    LAST_DB_UPDATE_FMT_STR,
)
from pipictureframe.picdb.DbObjects import Metadata
from pipictureframe.picdb.PictureUpdater import PictureData

FILTER_RATING_BELOW = "frb"
FILTER_RATING_ABOVE = "fra"

log = logging.getLogger(__name__)


class NextPictureManager:
    def __init__(self, config, pic_load_bg_proc: Process, db: Database):
        self.config = config
        self.pic_load_bg_proc = pic_load_bg_proc
        self.filters = dict()
        if config.min_rating:
            self.filters[FILTER_RATING_BELOW] = config.min_rating
        if config.max_rating:
            self.filters[FILTER_RATING_ABOVE] = config.max_rating

        self.db = db
        self.last_db_update = datetime(1, 1, 1)

        self.cur_pic_num = 0
        self.sample_list = None
        self.reload_pictures()

    def reload_pictures(self):
        self.cur_pic_num = 0
        self.sample_list = self._load_pictures()

    def get_next_picture(self):
        if self.cur_pic_num >= len(self.sample_list):
            self.reload_pictures()
        cur_pic = self.sample_list[self.cur_pic_num]
        self._inc_times_shown(cur_pic)
        self.cur_pic_num += 1
        return cur_pic

    def _inc_times_shown(self, cur_pic):
        session = self.db.get_session()
        cur_pic.times_shown += 1
        session.merge(cur_pic)
        session.commit()
        session.close()

    def _load_pictures(self) -> List[PictureData]:
        full_picture_list = self._load_pictures_from_db()
        if full_picture_list is None:
            log.debug("No update in db detected. Keeping current picture list.")
            return self.sample_list
        if len(full_picture_list) == 0:
            if self.pic_load_bg_proc.is_alive():
                time.sleep(
                    10
                )  # Give the bg process some time to load pictures into the database
                full_picture_list = self._load_pictures()
            else:
                log.fatal("No pictures found to display in database.")
                exit(1)

        sample_list = full_picture_list
        if not self.config.shuffle:
            sample_list.sort(key=lambda x: x.get_date_time())
        elif self.config.shuffle_weight == 0:
            random.shuffle(sample_list)
        else:
            nums_shown = [pic.times_shown for pic in full_picture_list]
            max_num_shown = max(nums_shown)
            weights = [
                (max_num_shown + 1 - x)
                / (max_num_shown + 1)
                * self.config.shuffle_weight
                + 1
                for x in nums_shown
            ]
            sample_list = random.choices(
                full_picture_list, weights=weights, k=len(full_picture_list)
            )
        return sample_list

    def _load_pictures_from_db(self) -> Optional[List[PictureData]]:
        session = self.db.get_session()
        last_db_update = datetime.strptime(
            session.query(Metadata)
            .filter(Metadata.key == LAST_DB_UPDATE_KEY_STR)
            .one()
            .value,
            LAST_DB_UPDATE_FMT_STR,
        )
        if self.last_db_update < last_db_update:
            q = session.query(PictureData)
            if FILTER_RATING_BELOW in self.filters:
                q.filter(PictureData.rating < self.filters[FILTER_RATING_BELOW])
            if FILTER_RATING_ABOVE in self.filters:
                q.filter(PictureData.rating > self.filters[FILTER_RATING_ABOVE])
            full_picture_list = q.all()
            log.info(f"{len(full_picture_list)} pictures loaded.")
            session.close()
            return full_picture_list
        else:
            return None



import logging
import os
from datetime import datetime

from sqlalchemy.orm import Session

from pipictureframe.picdb.Database import (
    Database,
    LAST_DB_UPDATE_KEY_STR,
    LAST_DB_UPDATE_FMT_STR,
)
from pipictureframe.picdb.DbObjects import PictureData, Metadata
from pipictureframe.utils.PictureReader import read_pictures_from_disk, PictureFile

log = logging.getLogger(__name__)


def update_pictures_in_db(pic_dir: str, connections_str: str):
    try:
        log.info(f"Starting update of db {connections_str} from directory {pic_dir}")
        pic_file_gen = read_pictures_from_disk(pic_dir)
        # Separate db instance created here since this runs a separate process
        db = Database(connections_str)

        session = db.get_session()
        db_changed = _clean_db(session)
        db_changed = _add_and_update_pics(pic_file_gen, session) or db_changed
        if db_changed:
            update_obj = Metadata(
                LAST_DB_UPDATE_KEY_STR, datetime.now().strftime(LAST_DB_UPDATE_FMT_STR)
            )
            session.merge(update_obj)
            session.commit()
        session.close()
    except Exception as e:
        log.fatal("Unexpected exception in picture update process.", exc_info=e)


def _clean_db(session) -> bool:
    all_pics = session.query(PictureData).all()
    num_deleted = 0
    for count, pic in enumerate(all_pics, 1):
        if not os.path.exists(pic.absolute_path):
            session.delete(pic)
            session.commit()
            num_deleted += 1
        if count % 1000 == 0:
            log.debug(f"Checked {count} db entries for missing hdd files.")
    log.info(f"Deleted {num_deleted} entries from db.")
    return num_deleted > 0


def _add_and_update_pics(pic_file_gen, session) -> bool:
    num_changed = 0
    for count, pic_file in enumerate(pic_file_gen, 1):
        try:
            pic_by_path = _get_pic_by_path(session, pic_file)
            # Is present
            if pic_by_path:
                # Has been modified
                if pic_by_path.mtime < pic_file.mtime:
                    log.debug(f"Updated timestamp detected for {pic_file.path}")
                    pic_data = PictureData.from_picture_file(pic_file)
                    pic_by_hash = _get_pic_by_hash(session, pic_data.hash_id)
                    # If modified but hash has not changed
                    if pic_by_hash:
                        log.debug(
                            f"Metadata but not hash has changed for {pic_file.path}"
                        )
                        session.merge(pic_data)
                    # If hash has changed
                    else:
                        log.debug(f"Hash has changed for {pic_file.path}")
                        session.delete(pic_by_path)
                        session.add(pic_data)
                else:
                    log.debug(f"File {pic_file.path} present in db and unchanged.")
                    continue
            # If not present
            else:
                pic_data = PictureData.from_picture_file(pic_file)
                pic_by_hash = _get_pic_by_hash(session, pic_data.hash_id)
                if pic_by_hash:
                    log.debug(
                        f"Picture has moved from {pic_by_hash.absolute_path} to {pic_data.absolute_path}"
                    )
                    session.merge(pic_data)
                else:
                    log.debug(f"Picture {pic_file.path} will be added to the database.")
                    session.add(pic_data)
            session.commit()
            num_changed += 1
            if count % 1000 == 0:
                log.debug(f"Checked {count} files for necessary db updates.")
        except Exception as e:
            log.error(
                f"Exception while trying to update db with {pic_file.path}.", exc_info=e
            )
            session.rollback()
    return num_changed > 0


def _get_pic_by_path(session: Session, pic_file: PictureFile) -> PictureData:
    q = session.query(PictureData).filter(PictureData.absolute_path == pic_file.path)
    return q.first()


def _get_pic_by_hash(session: Session, hash_id: str) -> PictureData:
    q = session.query(PictureData).filter(PictureData.hash_id == hash_id)
    return q.first()
